@@ -18,9 +18,7 @@
 //! another engineer approve and monitor any privileged sessions.
 
 // TODO: remove all to_string_lossy
-// TODO: switch from error_chain to failure crate?
 // TODO: error message when /var/run/sudo_pair missing
-// TODO: enable the ability to respond to `sudo --version`
 // TODO: iolog in `sudoreplay(8)` format
 // TODO: rustfmt
 // TODO: double-check all `as`-casts
@@ -31,8 +29,10 @@
 
 #![warn(bad_style)]
 #![warn(future_incompatible)]
+#![warn(nonstandard_style)]
 #![warn(rust_2018_compatibility)]
 #![warn(rust_2018_idioms)]
+#![warn(rustdoc)]
 #![warn(unused)]
 
 #![warn(bare_trait_objects)]
@@ -53,32 +53,15 @@
 // this entire crate is unsafe code
 #![allow(unsafe_code)]
 
-#![cfg_attr(feature="cargo-clippy", warn(clippy))]
-#![cfg_attr(feature="cargo-clippy", warn(clippy_complexity))]
-#![cfg_attr(feature="cargo-clippy", warn(clippy_correctness))]
-#![cfg_attr(feature="cargo-clippy", warn(clippy_pedantic))]
-#![cfg_attr(feature="cargo-clippy", warn(clippy_perf))]
-#![cfg_attr(feature="cargo-clippy", warn(clippy_style))]
-
-// TODO: we can remove `bindgen` as a direct dependency and just bundle
-// its output since it's static; these should pass much more reliably
-// then
-//
-// #![cfg_attr(feature="cargo-clippy", warn(clippy_cargo))]
-
-extern crate libc;
-
-#[macro_use]
-extern crate error_chain;
-
-#[macro_use]
-extern crate sudo_plugin;
+#![cfg_attr(feature="cargo-clippy", warn(clippy::all))]
 
+mod errors;
 mod template;
 mod socket;
 
-use template::Spec;
-use socket::Socket;
+use crate::errors::*;
+use crate::template::Spec;
+use crate::socket::Socket;
 
 use std::collections::HashSet;
 use std::fs::File;
@@ -88,8 +71,9 @@ use std::path::PathBuf;
 
 use libc::{gid_t, mode_t, uid_t};
 
-use sudo_plugin::errors::*;
-use sudo_plugin::OptionMap;
+use failure::ResultExt;
+
+use sudo_plugin::*;
 
 const DEFAULT_BINARY_PATH      : &str       = "/usr/bin/sudo_approve";
 const DEFAULT_USER_PROMPT_PATH : &str       = "/etc/sudo_pair.prompt.user";
@@ -111,32 +95,28 @@ sudo_io_plugin! {
 }
 
 struct SudoPair {
-    plugin:  &'static sudo_plugin::Plugin,
+    plugin:  &'static Plugin,
     options: PluginOptions,
-    socket:  Option<Socket>
+    socket:  Option<Socket>,
 }
 
 impl SudoPair {
-    fn open(plugin: &'static sudo_plugin::Plugin) -> Result<Self> {
+    fn open(plugin: &'static Plugin) -> Result<Self> {
         // TODO: convert all outgoing errors to be unauthorized errors
-        let options = PluginOptions::from(&plugin.plugin_options);
-
         let mut pair = Self {
             plugin,
-            options,
-            socket: None,
+            options: PluginOptions::from(&plugin.plugin_options),
+            socket:  None,
         };
 
-        if pair.is_sudoing_to_user_and_group() {
-            bail!(ErrorKind::Unauthorized(
-                "sudo_pair doesn't support sudoing to both a user and a group".into()
-            ));
-        }
-
         if pair.is_exempt() {
             return Ok(pair)
         }
 
+        if pair.is_sudoing_to_user_and_group() {
+            return Err(ErrorKind::SudoToUserAndGroup.into());
+        }
+
         let template_spec = pair.template_spec();
 
         pair.local_pair_prompt(&template_spec);
@@ -173,7 +153,7 @@ impl SudoPair {
 
     fn close(&mut self, _: i64, _: i64) {
         // if we have a socket, close it
-        let _ = self.socket.as_mut().map(|s| s.close());
+        let _ = self.socket.as_mut().map(Socket::close);
     }
 
     fn log_ttyout(&mut self, log: &[u8]) -> Result<()> {
@@ -203,12 +183,10 @@ impl SudoPair {
     fn log_output(&mut self, log: &[u8]) -> Result<()> {
         // if we have a socket, write to it
         self.socket.as_mut().map_or(Ok(()), |socket| {
-            socket
-                .write_all(log)
-                .chain_err(|| ErrorKind::Unauthorized(
-                    "pair terminated the session".into()
-                ))
-        })
+            socket.write_all(log)
+        }).context(ErrorKind::SessionTerminated)?;
+
+        Ok(())
     }
 
     fn log_disabled(&mut self, _: &[u8]) -> Result<()> {
@@ -217,9 +195,7 @@ impl SudoPair {
             return Ok(());
         }
 
-        bail!(ErrorKind::Unauthorized(
-            "redirection of stdin to paired sessions is prohibited".into()
-        ));
+        Err(ErrorKind::StdinRedirected.into())
     }
 
     fn local_pair_prompt(&self, template_spec: &Spec) {
@@ -275,7 +251,7 @@ impl SudoPair {
             self.socket_uid(),
             self.socket_gid(),
             self.socket_mode(),
-        ).chain_err(|| ErrorKind::Unauthorized("unable to connect to a pair".into()))?;
+        ).context(ErrorKind::CommunicationError)?;
 
         self.socket = Some(socket);
 
@@ -293,14 +269,13 @@ impl SudoPair {
 
         let socket = self.socket
             .as_mut()
-            .ok_or_else(|| ErrorKind::Unauthorized("unable to connect to a pair".into()))?;
+            .ok_or(ErrorKind::CommunicationError)?;
 
         socket.write_all(&prompt[..])
-            .chain_err(|| ErrorKind::Unauthorized("unable to ask pair for approval".into()))?;
+            .context(ErrorKind::CommunicationError)?;
 
-        // ensure the entire prompt was written to the pair
         socket.flush()
-            .chain_err(|| ErrorKind::Unauthorized("unable to ask pair for approval".into()))?;
+            .context(ErrorKind::CommunicationError)?;
 
         // default `response` to something other than success, since
         // `read` might return without actually having written anything;
@@ -314,7 +289,7 @@ impl SudoPair {
         // value because if the read was successful, we're guaranteed to
         // have read at least one byte
         let _ = socket.read(&mut response)
-            .chain_err(|| ErrorKind::Unauthorized("denied by pair".into()))?;
+            .context(ErrorKind::SessionDeclined)?;
 
         // echo back out the response, since the client is anticipated
         // to be noecho
@@ -323,7 +298,7 @@ impl SudoPair {
 
         match &response {
             b"y" | b"Y" => Ok(()),
-            _           => Err(ErrorKind::Unauthorized("denied by pair".into()).into()),
+            _           => Err(ErrorKind::SessionDeclined.into()),
         }
     }
 
